/* Function prototypes */

void parse_username(char *username, char **uname, char **room, char **wp,
		    char **hp, char **other);
int indent(int length);
void main_finger_part(char *username);
void show_configuration(void);
void read_configuration(void);
BOOL is_trusted(char *domain);
BOOL is_rejected(char *domain);
char *get_localhost(void);
BOOL search_illegal(char *str);
void display_file(char *disp_file);
BOOL no_finger(char *username);
void show_version_information(void);
void initialize_daemon(void);
void strmcpy(char **t, const char *s);
void int_handler(int signo);
int is_fake(char *user);
void become_nobody(void);
/* void lowercase(char **t, const char *s); */
char *lowercase(const char *s);
int is_searchable(char *user);
void call_userscript(char *user, char *realuser);
void show_services(void);
void pprintf(char *string);
BOOL check_for_user(char *username);
void check_illegal(char *username);
